#include "DataFormats/Common/interface/Handle.h"
#include "DataFormats/FEDRawData/interface/FEDNumbering.h"
#include "DataFormats/FEDRawData/interface/FEDRawData.h"
#include "DataFormats/FEDRawData/interface/FEDRawDataCollection.h"
#include "DataFormats/Scalers/interface/DcsStatus.h"
#include "DataFormats/OnlineMetaData/interface/DCSRecord.h"
#include "FWCore/Framework/interface/Event.h"
#include "FWCore/Framework/interface/stream/EDFilter.h"
#include "FWCore/MessageLogger/interface/MessageLogger.h"
#include "FWCore/ParameterSet/interface/ParameterSet.h"
#include "FWCore/ParameterSet/interface/ConfigurationDescriptions.h"
#include "FWCore/ParameterSet/interface/ParameterSetDescription.h"

class DetectorStateFilter : public edm::stream::EDFilter<> {
public:
  DetectorStateFilter(const edm::ParameterSet&);
  ~DetectorStateFilter() override;

  static void fillDescriptions(edm::ConfigurationDescriptions& descriptions);

private:
  bool filter(edm::Event&, edm::EventSetup const&) override;

  const bool verbose_;
  uint64_t nEvents_, nSelectedEvents_;
  bool detectorOn_;
  const std::string detectorType_;
  const edm::EDGetTokenT<DcsStatusCollection> dcsStatusLabel_;
  const edm::EDGetTokenT<DCSRecord> dcsRecordToken_;

  bool checkDCSStatus(const DcsStatusCollection& dcsStatus);
  bool checkDCSRecord(const DCSRecord& dcsRecod);
};

//
// -- Constructor
//
DetectorStateFilter::DetectorStateFilter(const edm::ParameterSet& pset)
    : verbose_(pset.getUntrackedParameter<bool>("DebugOn", false)),
      detectorType_(pset.getUntrackedParameter<std::string>("DetectorType", "sistrip")),
      dcsStatusLabel_(consumes<DcsStatusCollection>(
          pset.getUntrackedParameter<edm::InputTag>("DcsStatusLabel", edm::InputTag("scalersRawToDigi")))),
      dcsRecordToken_(consumes<DCSRecord>(
          pset.getUntrackedParameter<edm::InputTag>("DCSRecordLabel", edm::InputTag("onlineMetaDataDigis")))) {
  nEvents_ = 0;
  nSelectedEvents_ = 0;
  detectorOn_ = false;
}

//
// -- Destructor
//
DetectorStateFilter::~DetectorStateFilter() = default;

//*********************************************************************//
bool DetectorStateFilter::checkDCSStatus(const DcsStatusCollection& dcsStatus)
//*********************************************************************//
{
  bool accepted = false;
  if (detectorType_ == "pixel") {
    if ((dcsStatus)[0].ready(DcsStatus::BPIX) && (dcsStatus)[0].ready(DcsStatus::FPIX)) {
      accepted = true;
      nSelectedEvents_++;
    } else {
      accepted = false;
    }
    //if (verbose_)
    edm::LogInfo("DetectorStatusFilter") << " Total Events " << nEvents_ << " Selected Events " << nSelectedEvents_
                                         << " DCS States : "
                                         << " BPix " << (dcsStatus)[0].ready(DcsStatus::BPIX) << " FPix "
                                         << (dcsStatus)[0].ready(DcsStatus::FPIX) << " Detector State " << accepted
                                         << std::endl;
  } else if (detectorType_ == "sistrip") {
    if ((dcsStatus)[0].ready(DcsStatus::TIBTID) && (dcsStatus)[0].ready(DcsStatus::TOB) &&
        (dcsStatus)[0].ready(DcsStatus::TECp) && (dcsStatus)[0].ready(DcsStatus::TECm)) {
      accepted = true;
      nSelectedEvents_++;
    } else {
      accepted = false;
    }
    //if (verbose_)
    edm::LogInfo("DetectorStatusFilter") << " Total Events " << nEvents_ << " Selected Events " << nSelectedEvents_
                                         << " DCS States : "
                                         << " TEC- " << (dcsStatus)[0].ready(DcsStatus::TECm) << " TEC+ "
                                         << (dcsStatus)[0].ready(DcsStatus::TECp) << " TIB/TID "
                                         << (dcsStatus)[0].ready(DcsStatus::TIBTID) << " TOB "
                                         << (dcsStatus)[0].ready(DcsStatus::TOB) << " Detector States " << accepted
                                         << std::endl;
  }
  return accepted;
}

//*********************************************************************//
bool DetectorStateFilter::checkDCSRecord(const DCSRecord& dcsRecord)
//*********************************************************************//
{
  bool accepted = false;
  if (detectorType_ == "pixel") {
    if (dcsRecord.highVoltageReady(DCSRecord::Partition::BPIX) &&
        dcsRecord.highVoltageReady(DCSRecord::Partition::FPIX)) {
      accepted = true;
      nSelectedEvents_++;
    } else {
      accepted = false;
    }
    //if (verbose_)
    edm::LogInfo("DetectorStatusFilter") << " Total Events " << nEvents_ << " Selected Events " << nSelectedEvents_
                                         << " DCS States : "
                                         << " BPix " << dcsRecord.highVoltageReady(DCSRecord::Partition::BPIX)
                                         << " FPix " << dcsRecord.highVoltageReady(DCSRecord::Partition::FPIX)
                                         << " Detector State " << accepted << std::endl;
  } else if (detectorType_ == "sistrip") {
    if (dcsRecord.highVoltageReady(DCSRecord::Partition::TIBTID) &&
        dcsRecord.highVoltageReady(DCSRecord::Partition::TOB) &&
        dcsRecord.highVoltageReady(DCSRecord::Partition::TECp) &&
        dcsRecord.highVoltageReady(DCSRecord::Partition::TECm)) {
      accepted = true;
      nSelectedEvents_++;
    } else {
      accepted = false;
    }
    //if (verbose_)
    edm::LogInfo("DetectorStatusFilter") << " Total Events " << nEvents_ << " Selected Events " << nSelectedEvents_
                                         << " DCS States : "
                                         << " TEC- " << dcsRecord.highVoltageReady(DCSRecord::Partition::TECm)
                                         << " TEC+ " << dcsRecord.highVoltageReady(DCSRecord::Partition::TECp)
                                         << " TIB/TID " << dcsRecord.highVoltageReady(DCSRecord::Partition::TIBTID)
                                         << " TOB " << dcsRecord.highVoltageReady(DCSRecord::Partition::TOB)
                                         << " Detector States " << accepted << std::endl;
  }
  return accepted;
}

//*********************************************************************//
bool DetectorStateFilter::filter(edm::Event& evt, edm::EventSetup const& es)
//*********************************************************************//
{
  nEvents_++;
  // Check Detector state Only for Real Data and return true for MC
  if (evt.isRealData()) {
    edm::Handle<DcsStatusCollection> dcsStatus;
    evt.getByToken(dcsStatusLabel_, dcsStatus);
    edm::Handle<DCSRecord> dcsRecord;
    evt.getByToken(dcsRecordToken_, dcsRecord);

    if (dcsStatus.isValid() && !dcsStatus->empty()) {
      // if the old style DCS status is valid (Run1 + Run2)
      detectorOn_ = checkDCSStatus(*dcsStatus);
    } else if (dcsRecord.isValid()) {
      // in case of real data check for DCSRecord content (Run >=3)
      detectorOn_ = checkDCSRecord(*dcsRecord);
    } else {
      edm::LogError("DetectorStatusFilter")
          << "Error! can't get the products, neither DCSRecord, nor scalersRawToDigi: accept in any case!";
      detectorOn_ = true;
    }
  } else {
    detectorOn_ = true;
    nSelectedEvents_++;
    //if (verbose_)
    edm::LogInfo("DetectorStatusFilter") << "Total MC Events " << nEvents_ << " Selected Events " << nSelectedEvents_
                                         << " Detector States " << detectorOn_ << std::endl;
  }
  return detectorOn_;
}

// ------------ method fills 'descriptions' with the allowed parameters for the module  ------------
void DetectorStateFilter::fillDescriptions(edm::ConfigurationDescriptions& descriptions) {
  edm::ParameterSetDescription desc;
  desc.setComment("filters on the HV status of the Tracker (either pixels or strips)");
  desc.addUntracked<bool>("DebugOn", false)->setComment("activates debugging");
  desc.addUntracked<std::string>("DetectorType", "sistrip")->setComment("either strips or pixels");
  desc.addUntracked<edm::InputTag>("DcsStatusLabel", edm::InputTag("scalersRawToDigi"))
      ->setComment("event data for DCS (Run2)");
  desc.addUntracked<edm::InputTag>("DCSRecordLabel", edm::InputTag("onlineMetaDataDigis"))
      ->setComment("event data for DCS (Run3)");
  descriptions.add("_detectorStateFilter", desc);
}

#include "FWCore/Framework/interface/MakerMacros.h"
DEFINE_FWK_MODULE(DetectorStateFilter);
